package common.matroska;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Map;
import java.util.TimeZone;

import common.MuxBuffer;

public class MatroskaDecoder extends MuxBuffer implements MatroskaConstants {
	private static ByteBuffer idBuffer=ByteBuffer.allocateDirect(12);
	public static final Calendar MATROSKA_EPOCH=Calendar.getInstance(TimeZone.getTimeZone("GMT"));

	private static ByteBuffer intBuffer=ByteBuffer.allocateDirect(8);
	private static MatroskaDecoder intDecoder=new MatroskaDecoder(intBuffer);
	
	static final Map<Integer, MClass> idClassMap=new HashMap<Integer, MClass>();
	
	public static class MClass {
		public int id;
		public String name;
		public int type;
		
		private static final String[] TYPE_NAMES={"Object","u-int","string", "binary", "utf-8", "float", "date", "int"};

		public static final int TYPE_MASTER=0;
		public static final int TYPE_UINT=1;
		public static final int TYPE_STRING=2;
		public static final int TYPE_BINARY=3;
		public static final int TYPE_UTF8=4;
		public static final int TYPE_FLOAT=5;
		public static final int TYPE_DATE=6;
		public static final int TYPE_SINT=7;
		
		
		public static final int TYPE_SIMPLE_BLOCK=100;
		public static final int TYPE_TIMECODE=101;
		
		MClass(int id, String name, int type) {
			this.id=id;
			this.name=name;
			this.type=type;
		}
		
		public String toString() {
			return "0x" + Integer.toString(id, 16) + " " + name + ":" + TYPE_NAMES[type];
		}
	}
	
	
	private static void addClass(int id, String name, int type) {
		MClass mClass=new MClass(id, name, type);
		idClassMap.put(Integer.valueOf(id), mClass);
	}
	static {
		MATROSKA_EPOCH.clear();
		MATROSKA_EPOCH.set(2001, 0, 1);
		//System.out.println(MATROSKA_EPOCH);
		
		addClass(EBML, "EBML", MClass.TYPE_MASTER);
		addClass(EBMLVersion, "EBMLVersion", MClass.TYPE_UINT);
		addClass(EBMLReadVersion, "EBMLReadVersion", MClass.TYPE_UINT);
		addClass(EBMLMaxIDLength, "EBMLMaxIDLength", MClass.TYPE_UINT);
		addClass(EBMLMaxSizeLength, "EMBLMaxSizeLength", MClass.TYPE_UINT);
		addClass(DocType, "DocType", MClass.TYPE_STRING);
		addClass(DocTypeVersion, "DocTypeVersion", MClass.TYPE_UINT);
		addClass(DocTypeReadVersion, "DocTypeReadVersion", MClass.TYPE_UINT);
		
		addClass(CRC_32, "CRC-32", MClass.TYPE_BINARY);
		addClass(Void, "Void", MClass.TYPE_BINARY);
		
		addClass(Segment, "Segment", MClass.TYPE_MASTER);
		
		addClass(SeekHead, "SeekHead", MClass.TYPE_MASTER);
		addClass(Seek, "Seek", MClass.TYPE_MASTER);
		addClass(SeekID, "SeekID", MClass.TYPE_BINARY);
		addClass(SeekPosition, "SeekPosition", MClass.TYPE_UINT);
		
		addClass(Info, "Info", MClass.TYPE_MASTER);
		addClass(SegmentUID, "SegmentUID", MClass.TYPE_BINARY);
		addClass(TimecodeScale,"TimecodeScale", MClass.TYPE_UINT);
		addClass(Duration,"Duration", MClass.TYPE_FLOAT);
		addClass(DateUTC, "DateUTC", MClass.TYPE_DATE);
		addClass(MuxingApp, "MuxingApp", MClass.TYPE_UTF8);
		addClass(WritingApp, "WritingApp", MClass.TYPE_UTF8);
		addClass(Title, "Title", MClass.TYPE_UTF8);
		addClass(Tracks, "Track", MClass.TYPE_MASTER);
		addClass(TrackEntry, "TrackEntry", MClass.TYPE_MASTER);
		addClass(TrackNumber, "TrackNumber", MClass.TYPE_UINT);
		addClass(TrackUID, "TrackUID", MClass.TYPE_UINT);
		addClass(TrackType, "TrackType", MClass.TYPE_UINT);
		addClass(FlagLacing, "FlagLacing", MClass.TYPE_UINT);
		addClass(FlagDefault, "FlagDefault", MClass.TYPE_UINT);
		addClass(CodecID, "CodecID", MClass.TYPE_STRING);
		addClass(CodecPrivate, "CodecPrivate", MClass.TYPE_BINARY);
		addClass(DefaultDuration, "DefaultDuration", MClass.TYPE_UINT);
		addClass(Name, "Name", MClass.TYPE_UTF8);
		addClass(Language, "Language", MClass.TYPE_STRING);
		addClass(FlagEnabled, "FlagEnabled", MClass.TYPE_UINT);
		addClass(FlagForced, "FlagForced", MClass.TYPE_UINT);
		addClass(MinCache, "MinCache", MClass.TYPE_UINT);
		addClass(TrackTimecodeScale, "TrackTimecodeScale", MClass.TYPE_FLOAT);
		addClass(MaxBlockAdditionID, "MaxBlockAdditionID", MClass.TYPE_UINT);
		addClass(CodecDecodeAll, "CodecDecodeAll", MClass.TYPE_UINT);
		
		addClass(Video, "Video", MClass.TYPE_MASTER);
		addClass(PixelWidth, "PixelWidth", MClass.TYPE_UINT);
		addClass(PixelHeight, "PixelHeight", MClass.TYPE_UINT);
		addClass(DisplayWidth, "DisplayWidth", MClass.TYPE_UINT);
		addClass(DisplayHeight, "DisplayHeight", MClass.TYPE_UINT);
		addClass(DisplayUnit, "DisplayUnit", MClass.TYPE_UINT);
		addClass(FlagInterlaced, "FlagInterlaced", MClass.TYPE_UINT);

		addClass(Audio, "Audio", MClass.TYPE_MASTER);
		addClass(SamplingFrequency, "SamplingFrequency", MClass.TYPE_UINT);
		addClass(Channels, "Channels", MClass.TYPE_UINT);
		
		addClass(ContentEncodings, "ContentEncodings", MClass.TYPE_MASTER);
		addClass(ContentEncoding, "ContentEncoding", MClass.TYPE_MASTER);
		addClass(ContentCompression, "ContentCompression", MClass.TYPE_MASTER);
		addClass(ContentCompAlgo, "ContentCompAlgo", MClass.TYPE_UINT);
		addClass(ContentCompSettings, "ContentCompSettings", MClass.TYPE_BINARY);
	
		addClass(Cluster, "Cluster", MClass.TYPE_MASTER);
		//addClass(Timecode, "Timecode", MClass.TYPE_UINT);
		addClass(Timecode, "Timecode", MClass.TYPE_TIMECODE);
		addClass(BlockGroup, "BlockGroup", MClass.TYPE_MASTER);
		addClass(Block,"Block", MClass.TYPE_BINARY);
		addClass(ReferenceBlock, "ReferenceBlock", MClass.TYPE_UINT); //Technically this is S-INT
		addClass(SimpleBlock, "SimpleBlock", MClass.TYPE_SIMPLE_BLOCK);
		//addClass(SimpleBlock, "SimpleBlock", MClass.TYPE_BINARY);

		addClass(Chapters, "Chapters", MClass.TYPE_MASTER);
		addClass(EditionEntry, "EditionEntry", MClass.TYPE_MASTER);
		addClass(EditionFlagHidden, "EditionFlagHidden", MClass.TYPE_UINT);
		addClass(EditionFlagDefault, "EditionFlagDefault", MClass.TYPE_UINT);
		addClass(EditionUID, "EditionUID", MClass.TYPE_UINT);
		addClass(ChapterAtom, "ChapterAtom", MClass.TYPE_MASTER);
		addClass(ChapterUID, "ChapterUID", MClass.TYPE_UINT);
		addClass(ChapterTimeStart, "ChapterTimeStart", MClass.TYPE_UINT);
		addClass(ChapterFlagHidden, "ChapterFlagHidden", MClass.TYPE_UINT);
		addClass(ChapterFlagEnabled, "ChapterFlagEnabled", MClass.TYPE_UINT);
		addClass(ChapterDisplay, "ChapterDisplay", MClass.TYPE_MASTER);
		addClass(ChapString, "ChapString", MClass.TYPE_UTF8);
		addClass(ChapLanguage, "ChapLanguage", MClass.TYPE_STRING);
		
		addClass(Cues, "Cues", MClass.TYPE_MASTER);
		addClass(CuePoint, "CuePoint", MClass.TYPE_MASTER);
		addClass(CueTime, "CueTime", MClass.TYPE_UINT);
		addClass(CueTrackPositions, "CueTrackPositions", MClass.TYPE_MASTER);
		addClass(CueTrack, "CueTrack", MClass.TYPE_UINT);
		addClass(CueClusterPosition, "CueClusterPosition", MClass.TYPE_UINT);
		
	}
	
	public static MClass getMClass(int id) {
		MClass mClass=idClassMap.get(id);
		if (mClass==null) {
			System.out.println("Unknown matroska class: 0x" + Integer.toString(id, 16));
		}
		return mClass;
	}
	public static class IdInfo {
		public int id;
		public long len;
		public int idLen;
		
		public IdInfo(int id, long len, int idLen) {
			this.id=id;
			this.len=len;
			this.idLen=idLen;
		}
		public String toString() {
			return "0x:" + Integer.toString(id, 16) + " 0x" + Long.toString(len, 16);
		}
	}
	public MatroskaDecoder(ByteBuffer buffer) {
		super(buffer);
	}
		
	public int getIdValue() {
		int b=buffer.get() & 0xff;
		
		if (b >= 0x80) {
			return b & 0xff;
		} else if (b >= 0x40) {
			return ((b & 0x7f) << 8) | (buffer.get() & 0xff);
		} else if (b >= 0x20) {
			return ((b & 0x3f) << 16) | (buffer.getShort() & 0xffff);
		} else if (b >= 0x10) {
			buffer.position(buffer.position()-1);
			return buffer.getInt() & 0x1fffffff;
		} else {
			throw new RuntimeException("Entity larger than 4 bytes");
		}
		
	}
	public long getULong(int i) {
		switch (i) {
		case 1:
			return buffer.get() & 0xffl;
		case 2:
			return buffer.getShort() & 0xffffl;
		case 4:
			return buffer.getInt() & 0xffffffffl;
		case 8:
			return buffer.getLong(); // Signed
		case 3:
			return ((buffer.get() & 0xff) << 16) | (buffer.getShort() & 0xffff);
		case 5:
			return ((buffer.get() & 0xffl) << 32) | (buffer.getInt() & 0xffffffffl);
		case 6:
			return ((buffer.getShort() & 0xffffl) << 32) | (buffer.getInt() & 0xffffffffl);
		case 7:
			return ((buffer.get() & 0xffl) << 48) | ((buffer.getShort() & 0xffffl) << 32) | (buffer.getInt() & 0xffffffffl);  
		default:
			throw new RuntimeException("Unsupported len: " + i);
		}
	}
	public long getSLong(int i) {
		switch (i) {
		case 1:
			return buffer.get();
		case 2:
			return buffer.getShort();
		case 4:
			return buffer.getInt();
		case 8:
			return buffer.getLong();
		case 3:
			return (buffer.get() << 16) | buffer.getShort();
		case 5:
			return (((long)buffer.get()) << 32) | (buffer.getInt() & 0xffffffffl);
		case 6:
			return (((long)buffer.getShort()) << 32) | (buffer.getInt() & 0xffffffffl);
		case 7:
			return (((long)buffer.get()) & 0xffl << 48) | ((buffer.getShort() & 0xffffl) << 32) | (buffer.getInt() & 0xffffffffl);  
		default:
			throw new RuntimeException("Unsupported len: " + i);
		}
	}
	
	
	public long getSize() {
		int b=buffer.get() & 0xff;
		long l=b;
		long mask=0x7f;
		while (b < 0x80) {
			b<<=1;
			mask=(mask<<7) | 0x7f;
			l=(l<<=8) + (buffer.get() & 0xff);
		}
		return l & mask;
	}
	/**
	 * Returns an int array
	 * 0=id
	 * 1=segment len
	 * 2=header len
	 */
	public static IdInfo getNextChunkInfo(FileChannel channel) throws IOException {
		idBuffer.position(0);
		channel.read(idBuffer);
		idBuffer.flip();
		MatroskaDecoder decoder=new MatroskaDecoder(idBuffer);
		channel.position(channel.position()-idBuffer.limit());
		IdInfo idInfo=new IdInfo(decoder.getIdValue(),decoder.getSize(),idBuffer.position());
		MClass mClass=idClassMap.get(idInfo.id);
		if (mClass==null) {
			System.err.println("Unsupported Matroska Type: 0x" + Integer.toString(idInfo.id, 16));
			addClass(idInfo.id, "Unsupported", MClass.TYPE_BINARY);
		} /*else {
			System.out.println(mClass.name);
		}*/
		
		return idInfo;
	}
	
	public static IdInfo scanForId(FileChannel channel, int id, long end, boolean recurse) throws IOException {
		IdInfo idInfo=getNextChunkInfo(channel);
		if (idClassMap.get(idInfo.id).type==MClass.TYPE_MASTER && end==-1) {
			end=channel.position() + idInfo.len;
		}
		while (idInfo.id != id) {
			if (idClassMap.get(idInfo.id).type==MClass.TYPE_MASTER) {
				if (recurse) {
					channel.position(channel.position() + idInfo.idLen);
					idInfo=scanForId(channel, id, channel.position() + idInfo.len, recurse);
					if (idInfo!=null) {
						return idInfo;
					}
				} else {
					idInfo=null;
					break;
				}
			} else {
				channel.position(channel.position() + idInfo.idLen + idInfo.len);
			}
			if (channel.position() >= end) {
				idInfo=null;
				break;
			}
			idInfo=getNextChunkInfo(channel);
		}
		return idInfo;
	}
	public static synchronized long getULong(int len, FileChannel channel) throws IOException {
		intBuffer.limit(len);
		intBuffer.position(0);
		channel.read(intBuffer);
		intBuffer.rewind();
		return intDecoder.getULong(len);
		
	}
	/*public static long[] getNextChunkInfoLong(FileChannel channel) throws IOException {
		ByteBuffer buffer=ByteBuffer.allocateDirect(12);
		channel.read(buffer);
		buffer.flip();
		MatroskaDecoder decoder=new MatroskaDecoder(buffer);
		channel.position(channel.position()-buffer.limit());
		
		return new long[]{decoder.getValue(),decoder.getValueLong(),buffer.position()};
	}*/
	
}
