package muxer.filter;

import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;

import muxer.MuxFilter;
import muxer.MuxManager;
import muxer.MuxUtil;
import muxer.PacketBuffer;
import muxer.ts.TsRawMuxer;
import regulator.StreamPtsPacketCountRegulator;

import common.FileStream;
import common.Job;
import common.Logger;
import common.StreamDescriptor;
import common.mpeg.MpegConstants;
import common.mpeg.MpegUtil;
import common.packet.ClockReferenceI;
import common.packet.Packet;

import demux.Demuxer;
import demux.ts.M2tsRawDemuxer;

/**
 * Keeps a contiguous PCR/PTS/DTS timestamp stream 
 *
 */
public class ContiguousTimestampMuxFilter extends MuxFilter {
	private static final long PTS_TOLERANCE=MpegConstants.PTS_FREQ/4;// quarter second
	private static final long SCR_TOLERANCE=MpegConstants.PTS_FREQ/2;// half second
	private int streamIds[];
	//private int syncStreamIndex;
	private long lastStreamPtses[];
	private long lastScr=Long.MIN_VALUE;
	
	private StreamPtsPacketCountRegulator regulator;
	
	private long offset=0;
	private Logger logger=Logger.getLogger(this.getClass());
	
	public int getPosition() {
		return POSITION_POST_BUFFER;
	}
	protected void init() throws IOException {
		Map<Integer, StreamDescriptor> muxedStreamDescriptorMap=muxManager.getMuxedStreamDescriptorMap();
		ArrayList<StreamDescriptor> sdList=new ArrayList<StreamDescriptor>(muxedStreamDescriptorMap.size());
		for (Iterator<StreamDescriptor> i=muxedStreamDescriptorMap.values().iterator();i.hasNext();) {
			StreamDescriptor sd=i.next();
			if (sd.getClassification()==StreamDescriptor.CLASSIFICATION_AUDIO || sd.getClassification()==StreamDescriptor.CLASSIFICATION_VIDEO) {
				sdList.add(sd);
			}
		}

		int syncStreamIds[]=new int[sdList.size()];
		for (int i=0;i<syncStreamIds.length;i++) {
			syncStreamIds[i]=sdList.get(i).getStreamId();
		}
		streamIds=muxManager.getDemuxer().getStreamIds();
		lastStreamPtses=new long[streamIds.length];
		for (int i=0;i<lastStreamPtses.length;i++) {
			lastStreamPtses[i]=Long.MIN_VALUE;
		}
		//syncStreamIndex=getStreamIndex(syncStreamId);
		PacketBuffer packetBuffer=muxManager.createPacketBuffer();
		packetBuffer.addRegulator(regulator=new StreamPtsPacketCountRegulator(syncStreamIds, 5));
	}
	public int addPacket(Packet packet) throws IOException {
		long scr=packet.getClockReference(ClockReferenceI.CLOCK_REFERENCE_TYPE_SYSTEM);
		if (scr >=0 ) {
			if (lastScr != Long.MIN_VALUE && Math.abs(scr+offset-lastScr) > SCR_TOLERANCE) {
				//by the MPEG spec, the scr has to be delivered every .5 seconds
				calculateOffset("SCR", lastScr, scr);
			}
			lastScr=scr + offset;
			if (offset !=0) {
				packet.offset(offset);
			}
		}
		long pts=packet.getClockReference(ClockReferenceI.CLOCK_REFERENCE_TYPE_PTS);
		if (pts >=0) {
			int index=getStreamIndex(packet.getStreamId());
			if (lastStreamPtses[index]!=Long.MIN_VALUE) {
				
				if (Math.abs(pts + offset - lastStreamPtses[index]) > PTS_TOLERANCE) {
					calculateOffset("PTS", lastStreamPtses[index], pts);
				}
			}
			//Video can go backwards
			if (pts+offset > lastStreamPtses[index]) {
				lastStreamPtses[index]=pts+offset;
			}
			if (offset != 0) {
				packet.offset(offset);
			}
		}

		return super.addPacket(packet);
	}
	
	private void calculateOffset(String name, long lastTimestamp, long currentTimestamp) {
		if (logger.isDebug()) {
			logger.info("Large "+name+" jump " + MpegUtil.pesTimeToString(lastTimestamp) + " to " + MpegUtil.pesTimeToString(currentTimestamp) + " adjusting...");
		}
		int streamId=regulator.getLargestStream();
		if (logger.isDebug()) {
			logger.debug(regulator.toString());
			logger.debug("Using streamId: " + Integer.toString(streamId, 16) + " for sync");
		}
		long syncStreamPtsMode=MuxUtil.getMode(muxManager.getPacketBuffer().iterator(), streamId);
		if (logger.isDebug()) {
			logger.debug("PTS delta mode: " + MpegUtil.pesTimeToString(syncStreamPtsMode));
		}
		long nextSyncStreamPts=MuxUtil.getNextPacketPts(streamId, muxManager.getPacketBuffer());
		int syncStreamIndex=getStreamIndex(streamId);
		offset=lastStreamPtses[syncStreamIndex] + syncStreamPtsMode - nextSyncStreamPts;
		//offset=lastStreamPtses[1] + MpegConstants.PTS_FREQ*1001/24000 -nextSyncStreamPts; 
		
		for (int i=0;i<lastStreamPtses.length;i++) {
			logger.debug("0x" + Integer.toString(streamIds[i],16) + " Last: " + MpegUtil.pesTimeToString(lastStreamPtses[i]) + " Next: " + MpegUtil.pesTimeToString(MuxUtil.getNextPacketPts(streamIds[i], muxManager.getPacketBuffer())+offset));
		}
	}
	private final int getStreamIndex(int streamId) {
		for (int i=0;i<streamIds.length;i++) {
			if (streamIds[i]==streamId) {
				return i;
			}
		}
		return -1;
	}
	
	public static void main(String args[]) {
		Demuxer demuxer=new M2tsRawDemuxer();
		MuxManager muxManager=new MuxManager(demuxer);
		Job job=new Job(null);
		FileStream fileStream=new FileStream(job,FileStream.getFiles("I:\\BDMV\\STREAM\\00027.m2ts\nI:\\BDMV\\STREAM\\00028.m2ts\nI:\\BDMV\\STREAM\\00000.m2ts"));
		try {
			muxManager.getStreamDescriptorMap(fileStream);
			fileStream.reset();
			//TODO: Fix me
			//demuxer.setStreamIds(new int[]{0x1011,0x1101});
			muxManager.add(new ContiguousTimestampMuxFilter());
			muxManager.add(new TsRawMuxer(new FileOutputStream("H:\\Rat.ts").getChannel()));
			muxManager.mux(fileStream);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
}
